#!/usr/bin/env perl

use strict;
use warnings;

use File::Basename;
use File::Spec;
use File::Spec::Link;

my $reapr = File::Spec->rel2abs($0);
my $this_script = File::Spec::Link->resolve($0);
$this_script = File::Spec->rel2abs($this_script);
my ($scriptname, $scriptdir) = fileparse($this_script);
$scriptdir = File::Spec->rel2abs($scriptdir);
my $tabix = File::Spec->catfile($scriptdir, 'tabix/tabix');
my $bgzip = File::Spec->catfile($scriptdir, 'tabix/bgzip');
my $samtools = File::Spec->catfile($scriptdir, 'samtools');
my $version = '1.0.16';
my $ERROR_PREFIX = '[REAPR pipeline]';

if ($#ARGV == -1) {
    die qq/REAPR version: $version
Usage:
    reapr <task> [options]

Common tasks:
    facheck    - checks IDs in fasta file
    smaltmap   - map read pairs using SMALT: makes a BAM file to be used as
                 input to the pipeline
    perfectmap - make perfect uniquely mapping plot files
    pipeline   - runs the REAPR pipeline, using an assembly and mapped reads
                 as input, and optionally results of perfectmap.
                 (It runs facheck, preprocess, stats, fcdrate, score,
                 summary and break)
    plots      - makes Artemis plot files for a given contig, using results
                 from stats (and optionally results from score)
    seqrename  - renames all sequences in a BAM file: use this if you already
                 mapped your reads but then found facheck failed - saves
                 remapping the reads so that pipeline can be run

Advanced tasks:
    preprocess - preprocess files: necessary for running stats
    stats      - generates stats from a BAM file
    fcdrate    - estimates FCD cutoff for score, using results from stats
    score      - calculates scores and assembly errors, using results from stats
    summary    - make summary stats file, using results from score
    break      - makes broken assembly, using results from score
    gapresize  - experimental, calculates gap sizes based on read mapping
    perfectfrombam - generate perfect mapping plots from a bam file (alternative
                     to using perfectmap for large genomes)
/;
}

my %tasks = (
    'perfectmap' => "task_perfectmap.pl",
    'smaltmap' => "task_smaltmap.pl",
    'preprocess' => "task_preprocess.pl",
    'stats' => "task_stats",
    'break' => "task_break",
    'score' => "task_score",
    'plots' => "task_plots.pl",
    'pipeline'  => "x",
    'seqrename' => "x",
    'summary' => "task_summary.pl",
    'facheck' => 'task_facheck.pl',
    'gapresize' => 'task_gapresize',
    'fcdrate' => 'task_fcdrate',
    'perfectfrombam' => 'task_perfectfrombam.pl',
);

for my $k (keys %tasks) {
    $tasks{$k} = File::Spec->catfile($scriptdir, $tasks{$k});
}


if ($tasks{$ARGV[0]}) {
    if ($ARGV[0] eq "pipeline") {
        if ($#ARGV < 3 or $#ARGV > 4) {
            die qq/usage: reapr pipeline <assembly.fa> <in.bam> <out directory> [perfectmap prefix]

where 'perfectmap prefix' is optional and should be the prefix used when task
perfectmap was run.

It is assumed that reads in in.bam are 'innies', i.e. the correct orientation
is reads in a pair pointing towards each other (---> <---).
/;
        }
        my $ref = $ARGV[1];
        my $bam = $ARGV[2];
        my $dir = $ARGV[3];
        my $bash_script = "$dir.run-pipeline.sh";
        my $stats_prefix = '01.stats';
        my $fcdrate_prefix = '02.fcdrate';
        my $score_prefix = '03.score';
        my $break_prefix = '04.break';
        my $summary_prefix = '05.summary';

        my $perfect_prefix = "";
        if ($#ARGV == 4) {
            $perfect_prefix = File::Spec->rel2abs($ARGV[4]);
        }

        # make a bash script that runs all the pipeline commands
        my %commands;
        $commands{facheck} = "facheck $ref";
        $commands{preprocess} = "preprocess $ref $bam $dir\n"
                                . "cd $dir";

        if ($perfect_prefix) {
            $commands{stats} = "stats -p $perfect_prefix.perfect_cov.gz ./ $stats_prefix";
            $commands{score} = "score -P 5 00.assembly.fa.gaps.gz 00.in.bam $stats_prefix \$fcdcutoff $score_prefix";
        }
        else {
            $commands{stats} = "stats ./ $stats_prefix";
            $commands{score} = "score 00.assembly.fa.gaps.gz 00.in.bam $stats_prefix \$fcdcutoff $score_prefix";
        }

        $commands{fcdrate} = "fcdrate ./ $stats_prefix $fcdrate_prefix\n"
                              . "fcdcutoff=`tail -n 1 $fcdrate_prefix.info.txt | cut -f 1`";
        $commands{break} = "break 00.assembly.fa $score_prefix.errors.gff.gz $break_prefix";
        $commands{summary} = "summary 00.assembly.fa $score_prefix $break_prefix $summary_prefix";

        open F, ">$bash_script" or die "$ERROR_PREFIX Error opening file for writing '$bash_script'";
        print F "set -e\n"
                . "echo \"Running reapr version $version pipeline:\"\n"
                . "echo \"$reapr " . join(' ', @ARGV) . "\"\n\n";

        for my $task (qw/facheck preprocess stats fcdrate score break summary/) {
            print F "echo \"$ERROR_PREFIX Running $task\"\n"
                   . "$reapr $commands{$task}\n\n";
        }

        close F;

        exec "bash $bash_script" or die $!;
    }
    elsif ($ARGV[0] eq 'seqrename') {
        if ($#ARGV != 3) {
            die qq/usage: reapr seqrename <rename file> <in.bam> <out.bam>

where <rename file> is the file *.info made by 'facheck', which
contains the mapping of old name to new name
/;
        }

        # hash the old -> new names
        my %old2new;
        open F, $ARGV[1] or die $!;

        while (<F>) {
            chomp;
            my ($old, $new) = split /\t/;
            $old =~ s/\s+$//;
            $old2new{$old} = $new;
        }
        close F;
        $old2new{'*'} = '*';
        $old2new{'='} = '=';

        # make the new bam file
        open FIN, "$samtools view -h $ARGV[2]|" or die $!;
        open FOUT, "| $samtools view -bS - > $ARGV[3]" or die $!;

        while (<FIN>) {
            chomp;
            my @a  =split /\t/;

            if ($a[0] =~ /^\@SQ/) {
                $a[1] = "SN:" . $old2new{substr($a[1], 3)};
            }
            elsif ($a[0] !~ /^\@/) {
                $a[2] = $old2new{$a[2]};
                $a[6] = $old2new{$a[6]};
            }

            print FOUT join("\t", @a), "\n";
        }

        close FIN;
        close FOUT;
    }
    elsif ($#ARGV == 0) {
        print STDERR "usage:\nreapr $ARGV[0] ";
        exec "$tasks{$ARGV[0]} --wrapperhelp" or die;
    }
    else {
        my $cmd;

        if ($ARGV[0] eq "score") {
            my $score_out = "$ARGV[-1].per_base.gz";
            my $errors_out = "$ARGV[-1].errors.gff";
            $cmd = "$tasks{$ARGV[0]} " . join(" ", @ARGV[1..$#ARGV]) . "  | $bgzip -f -c > $score_out && $tabix -f -b 2 -e 2 $score_out && $bgzip -f $errors_out && $tabix -f -p gff $errors_out.gz";
        }
        else {
            $cmd = "$tasks{$ARGV[0]} " . join(" ", @ARGV[1..$#ARGV]);

            if ($ARGV[0] eq "stats") {
                my $outfile = "$ARGV[-1].per_base.gz";
                $cmd .= "  | $bgzip -f -c > $outfile && $tabix -f -b 2 -e 2 $outfile";
            }
        }

        exec $cmd or die;
    }
}
else {
    die qq/Task "$ARGV[0]" not recognised.\n/;
}
